from typing import Awaitable


async def play(
    sound_name: str, volume: int = 100, pitch: int = 0, pan: int = 0
) -> Awaitable:
    """Play a sound in the SPIKE App"""
    assert 0 <= volume <= 100, "Volume must be between 0 and 100"
    assert -100 <= pan <= 100, "Pan must be between -100 and 100"


def set_attributes(volume: int, pitch: int, pan: int) -> None:
    assert 0 <= volume <= 100, "Volume must be between 0 and 100"
    assert -100 <= pan <= 100, "Pan must be between -100 and 100"


def stop() -> None:
    pass



"""
To use a Motor add the following import statement to your project:

>>> import motor

All functions in the module should be called inside the motor module as a prefix like so:

>>> motor.run(port.A, 1000)
"""

from typing import Awaitable

READY = 0
RUNNING = 1
STALLED = 2
CANCELLED = 3
ERROR = 4
DISCONNECTED = 5

COAST = 0
BRAKE = 1
HOLD = 2
CONTINUE = 3
SMART_COAST = 4
SMART_BRAKE = 5

CLOCKWISE = 0
COUNTERCLOCKWISE = 1
SHORTEST_PATH = 2
LONGEST_PATH = 3


def absolute_position(port: int) -> int:
    """
    Get the absolute position of a Motor

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    """
    assert port in range(6), "Invalid port"


def get_duty_cycle(port: int) -> int:
    """
    Get the pwm of a Motor

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    """
    assert port in range(6), "Invalid port"


def relative_position(port: int) -> int:
    """
    Get the relative position of a Motor

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    """
    assert port in range(6), "Invalid port"


def reset_relative_position(port: int, position: int) -> None:
    """
    Reset the relative position of a Motor

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    position : int
        The degree of the motor
    """
    assert port in range(6), "Invalid port"
    assert isinstance(position, int), "Position must be an integer"


def run(port: int, velocity: int, *, acceleration: int = 1000) -> None:
    """
    Start a Motor at a constant speed

    Example
    --------
    ```python
    from hub import port
    import motor, time

    # Start motor
    motor.run(port.A, 1000)
    ```

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    velocity : int
        The velocity in degrees/sec

        Value ranges depends on motor type.

        - Small motor (essential): -660 to 660
        - Medium motor: -1110 to 1110
        - Large motor: -1050 to 1050

    acceleration : int
        The acceleration (deg/sec²) (1 - 10000)
    """
    assert port in range(6), "Invalid port"
    assert -10_000 <= velocity <= 10_000, "Invalid velocity"
    assert 1 <= acceleration <= 10_000, "Invalid acceleration"


async def run_for_degrees(
    port: int,
    degrees: int,
    velocity: int,
    *,
    stop: int = BRAKE,
    acceleration: int = 1000,
    deceleration: int = 1000,
) -> Awaitable:
    """
    Turn a motor for a specific number of degrees. When awaited returns a status of the movement that corresponds to one of the following constants:

    - `motor.READY`
    - `motor.RUNNING`
    - `motor.STALLED`
    - `motor.CANCELLED`
    - `motor.ERROR`
    - `motor.DISCONNECTED`

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    degrees : int
        The number of degrees
    velocity : int
        The velocity in degrees/sec

        Value ranges depends on motor type.

        - Small motor (essential): -660 to 660
        - Medium motor: -1110 to 1110
        - Large motor: -1050 to 1050
    stop : int
        The behavior of the Motor after it has stopped. Use the constants in the motor module.

        Possible values are
        - `motor.COAST` to make the motor coast until a stop
        - `motor.BRAKE` to brake and continue to brake after stop
        - `motor.HOLD` to tell the motor to hold it's position
        - `motor.CONTINUE` to tell the motor to keep running at whatever velocity it's running at until it gets another command
        - `motor.SMART_COAST` to make the motor brake until stop and then coast and compensate for inaccuracies in the next command
        - `motor.SMART_BRAKE` to make the motor brake and continue to brake after stop and compensate for inaccuracies in the next command
    acceleration : int
        The acceleration (deg/sec²) (1 - 10000)
    deceleration : int
        The deceleration (deg/sec²) (1 - 10000)
    """
    assert port in range(6), "Invalid port"
    assert isinstance(degrees, int), "Degrees must be an integer"
    assert -10_000 <= velocity <= 10_000, "Invalid velocity"
    assert stop in (COAST, BRAKE, HOLD, CONTINUE, SMART_COAST, SMART_BRAKE), (
        "Invalid stop behavior"
    )
    assert 1 <= acceleration <= 10_000, "Invalid acceleration"
    assert 1 <= deceleration <= 10_000, "Invalid deceleration"


async def run_for_time(
    port: int,
    duration: int,
    velocity: int,
    *,
    stop: int = BRAKE,
    acceleration: int = 1000,
    deceleration: int = 1000,
) -> Awaitable:
    """
    Run a Motor for a limited amount of time. When awaited returns a status of the movement that corresponds to one of the following constants:

    - `motor.READY`
    - `motor.RUNNING`
    - `motor.STALLED`
    - `motor.ERROR`
    - `motor.DISCONNECTED`

    Example
    -------
    ```python
    from hub import port
    import runloop
    import motor

    async def main():
        # Run at 1000 velocity for 1 second
        await motor.run_for_time(port.A, 1000, 1000)

        # Run at 280 velocity for 1 second
        await motor_pair.run_for_time(port.A, 1000, 280)

        # Run at 280 velocity for 10 seconds with a slow deceleration
        await motor_pair.run_for_time(port.A, 10000, 280, deceleration=10)

    runloop.run(main())
    ```

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    duration : int
        The duration in milliseconds
    velocity : int
        The velocity in degrees/sec

        Value ranges depends on motor type.

        - Small motor (essential): -660 to 660
        - Medium motor: -1110 to 1110
        - Large motor: -1050 to 1050
    direction : int
        The direction to turn.
        Options are:

        - `motor.CLOCKWISE`
        - `motor.COUNTERCLOCKWISE`
        - `motor.SHORTEST_PATH`
        - `motor.LONGEST_PATH`
    stop : int
        The behavior of the Motor after it has stopped. Use the constants in the motor module.

        Possible values are
        - `motor.COAST` to make the motor coast until a stop
        - `motor.BRAKE` to brake and continue to brake after stop
        - `motor.HOLD` to tell the motor to hold it's position
        - `motor.CONTINUE` to tell the motor to keep running at whatever velocity it's running at until it gets another command
        - `motor.SMART_COAST` to make the motor brake until stop and then coast and compensate for inaccuracies in the next command
        - `motor.SMART_BRAKE` to make the motor brake and continue to brake after stop and compensate for inaccuracies in the next command
    acceleration : int
        The acceleration (deg/sec²) (1 - 10000)
    deceleration : int
        The deceleration (deg/sec²) (1 - 10000)
    """
    assert port in range(6), "Invalid port"
    assert stop in (COAST, BRAKE, HOLD, CONTINUE, SMART_COAST, SMART_BRAKE), (
        "Invalid stop behavior"
    )
    assert -10_000 <= velocity <= 10_000, "Invalid velocity"
    assert 1 <= acceleration <= 10_000, "Invalid acceleration"
    assert 1 <= deceleration <= 10_000, "Invalid deceleration"


async def run_to_absolute_position(
    port: int,
    position: int,
    velocity: int,
    *,
    direction: int = SHORTEST_PATH,
    stop: int = BRAKE,
    acceleration: int = 1000,
    deceleration: int = 1000,
) -> Awaitable:
    """
    Turn a motor to an absolute position. When awaited returns a status of the movement that corresponds to one of the following constants:

    - `motor.READY`
    - `motor.RUNNING`
    - `motor.STALLED`
    - `motor.CANCELED`
    - `motor.ERROR`
    - `motor.DISCONNECTED`

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    position : int
        The target position in degrees
    velocity : int
        The velocity in degrees/sec

        Value ranges depends on motor type.

        - Small motor (essential): -660 to 660
        - Medium motor: -1110 to 1110
        - Large motor: -1050 to 1050
    direction : int
        The direction to turn the motor. Use the constants in the motor module.

        Possible values are
        - `motor.CLOCKWISE`
        - `motor.COUNTERCLOCKWISE`
        - `motor.SHORTEST_PATH`
        - `motor.LONGEST_PATH`
    stop : int
        The behavior of the Motor after it has stopped. Use the constants in the motor module.

        Possible values are
        - `motor.COAST` to make the motor coast until a stop
        - `motor.BRAKE` to brake and continue to brake after stop
        - `motor.HOLD` to tell the motor to hold it's position
        - `motor.CONTINUE` to tell the motor to keep running at whatever velocity it's running at until it gets another command
        - `motor.SMART_COAST` to make the motor brake until stop and then coast and compensate for inaccuracies in the next command
        - `motor.SMART_BRAKE` to make the motor brake and continue to brake after stop and compensate for inaccuracies in the next command
    acceleration : int
        The acceleration (deg/sec²) (1 - 10000)
    deceleration : int
        The deceleration (deg/sec²) (1 - 10000)
    """
    assert port in range(6), "Invalid port"
    assert direction in (
        CLOCKWISE,
        COUNTERCLOCKWISE,
        SHORTEST_PATH,
        LONGEST_PATH,
    ), "Invalid direction"
    assert stop in (COAST, BRAKE, HOLD, CONTINUE, SMART_COAST, SMART_BRAKE), (
        "Invalid stop behavior"
    )
    assert -10_000 <= velocity <= 10_000, "Invalid velocity"
    assert 1 <= acceleration <= 10_000, "Invalid acceleration"
    assert 1 <= deceleration <= 10_000, "Invalid deceleration"


def set_duty_cycle(port: int, pwm: int) -> None:
    """
    Set the duty cycle of a motor.

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    pwm : int
        The PWM value (-10000-10000)
    """
    assert port in range(6), "Invalid port"
    assert -10_000 <= pwm <= 10_000, "Invalid PWM"


def stop(port: int, *, stop: int = BRAKE) -> None:
    """
    Stop a motor.

    Example
    -------
    ```python
    from hub import port
    import motor, time

    # Start motor
    motor.run(port.A, 1000)

    # Wait for 2 seconds
    time.sleep_ms(2000)

    # Stop motor
    motor.stop(port.A)
    ```

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    stop : int
        The stop behavior. Use the constants in the motor module.

        Possible values are
        - `motor.COAST` to make the motor coast until a stop
        - `motor.BRAKE` to brake and continue to brake after stop
        - `motor.HOLD` to tell the motor to hold it's position
        - `motor.CONTINUE` to tell the motor to keep running at whatever velocity it's running at until it gets another command
        - `motor.SMART_COAST` to make the motor brake until stop and then coast and compensate for inaccuracies in the next command
        - `motor.SMART_BRAKE` to make the motor brake and continue to brake after stop and compensate for inaccuracies in the next command
    """
    assert port in range(6), "Invalid port"
    assert stop in (COAST, BRAKE, HOLD, CONTINUE, SMART_COAST, SMART_BRAKE), (
        "Invalid stop behavior"
    )


def velocity(port: int) -> int:
    """
    Get the velocity (deg/sec) of a Motor

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the `hub` module
    """
    assert port in range(6), "Invalid port"



"The bargraph module is used make bar graphs in the SPIKE App"

from typing import Awaitable


def change(color: int, value: float) -> None:
    assert 0 <= color <= 10, "Color must be between 0 and 10"
    assert 0 <= value <= 100, "Value must be between 0 and 100"


def clear_all() -> None:
    pass


async def get_value(color: int) -> Awaitable:
    pass


def set_value(color: int, value: float) -> None:
    pass


def show(fullscreen: bool) -> None:
    pass



def hardware_id() -> str:
    """Retrieve the hardware id."""
    pass


def power_off() -> int:
    """Turns off the hub."""
    pass


def temperature() -> int:
    """Retrieve the hub temperature. Measured in decidegrees celsius (d°C) which is 1 / 10 of a degree celsius (°C)"""
    pass



from enum import IntEnum, unique


@unique
class Color(IntEnum):
    BLACK = 0x00
    MAGENTA = 0x01
    PURPLE = 0x02
    BLUE = 0x03
    AZURE = 0x04
    TURQUOISE = 0x05
    GREEN = 0x06
    YELLOW = 0x07
    ORANGE = 0x08
    RED = 0x09
    WHITE = 0x0A
    NONE = 0xFF


@unique
class HubPort(IntEnum):
    A = 0x00
    B = 0x01
    C = 0x02
    D = 0x03
    E = 0x04
    F = 0x05


@unique
class HubFace(IntEnum):
    TOP = 0x00
    FRONT = 0x01
    RIGHT = 0x02
    BOTTOM = 0x03
    BACK = 0x04
    LEFT = 0x05


@unique
class ProgramAction(IntEnum):
    START = 0x00
    STOP = 0x01


@unique
class ResponseStatus(IntEnum):
    ACKNOWLEDGED = 0x00
    NOT_ACKNOWLEDGED = 0x01


@unique
class MotorEndState(IntEnum):
    """
    Smart coast/brake: methods of stopping a motor that compensate for inaccuracies when following commands.
    """

    COAST = 0x00
    BRAKE = 0x01
    HOLD = 0x02
    CONTINUE = 0x03
    COAST_SMART = 0x04
    BRAKE_SMART = 0x05
    DEFAULT = 0xFF


@unique
class MotorMoveDirection(IntEnum):
    CLOCKWISE = 0x00
    COUNTERCLOCKWISE = 0x01
    SHORTEST_PATH = 0x02
    LONGEST_PATH = 0x03


@unique
class MotorDeviceType(IntEnum):
    MEDIUM = 0x30
    LARGE = 0x31
    SMALL = 0x41



from binascii import crc32 as _crc32


def crc(data: bytes, seed=0, align=4):
    """
    Calculate the CRC32 of data with an optional seed and alignment.
    """
    remainder = len(data) % align
    if remainder:
        data += b"\x00" * (align - remainder)
    return _crc32(data, seed)



"""
The `force_sensor` module contains all functions and constants to use the Force Sensor.

To use the Force Sensor module add the following import statement to your project:

>>> import force_sensor

All functions in the module should be called inside the force_sensor module as a prefix like so:

>>> force_sensor.force(port.A)
"""


async def force(port: int) -> int:
    """
    Retrieves the measured force in decinewtons (0 to 100).

    Example
    -------
    >>> from hub import port
    >>> import force_sensor
    >>> print(force_sensor.force(port.A))

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the hub module.
    """
    assert port in range(6), "Invalid port"


async def pressed(port: int) -> bool:
    """
    Tests whether the button on the sensor is pressed. Returns true if the force sensor connected to port is pressed.

    Example
    -------
    >>> from hub import port
    >>> import force_sensor
    >>> print(force_sensor.pressed(port.A))

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the hub module.
    """
    assert port in range(6), "Invalid port"


async def raw(port: int) -> int:
    """
    Returns the raw, uncalibrated force value of the force sensor connected on `port`.

    Example
    -------
    >>> from hub import port
    >>> import force_sensor
    >>> print(force_sensor.raw(port.A))

    Parameters
    ----------
    port : int
        A port from the `port` submodule in the hub module.
    """
    assert port in range(6), "Invalid port"



import inspect, textwrap


def fn_to_string(fn) -> str:
    """Extract function body as a string without the function definition line."""
    lines, _ = inspect.getsourcelines(fn)  # full function source as lines
    body = textwrap.dedent("".join(lines[1:]))  # drop function name
    return body



"""
The `runloop` module contains all functions and constants to use the Runloop.

To use the Runloop module add the following import statement to your project:


>>> import runloop

All functions in the module should be called inside the `runloop` module as a prefix like so:

>>> runloop.run(some_async_function())
"""

from typing import Awaitable, Callable


def run(*functions: Awaitable) -> None:
    """
    Start any number of parallel `async` functions. This is the function you should use to create programs with a similar structure to Word Blocks.

    Parameters
    ----------
    functions : Awaitable
        The functions to run
    """
    pass


async def sleep_ms(duration: int) -> Awaitable:
    """
    Pause the execution of the application for any amount of milliseconds.

    Example
    -------
    ```python
    from hub import light_matrix
    import runloop

    async def main():
        light_matrix.write("Hi!")
        # Wait for ten seconds
        await runloop.sleep_ms(10000)
        light_matrix.write("Are you still here?")

    runloop.run(main())
    ```

    Parameters
    ----------
    duration : int
        The duration in milliseconds
    """
    pass


async def until(function: Callable[[], bool], timeout: int = 0) -> Awaitable:
    """
    Returns an awaitable that will return when the condition in the function or lambda passed is `True` or when it times out

    Example
    -------
    ```python
    import color_sensor
    import color
    from hub import port
    import runloop

    def is_color_red():
        return color_sensor.color(port.A) is color.RED

    async def main():
        # Wait until Color Sensor sees red
        await runloop.until(is_color_red)
        print("Red!")

    runloop.run(main())
    ```

    Parameters
    ----------
    function : Callable[[], bool]
        A callable with no parameters that returns either `True` or `False`. Callable is anything that can be called, so a `def` or a `lambda`
    timeout : int
        A timeout for the function in milliseconds. If the callable does not return `True` within the timeout, the `until` still resolves after the timeout. 0 means no timeout, in that case it will not resolve until the callable returns `True`
    """



"The display module is used show images in the SPIKE App"

IMAGE_ROBOT_1 = 1
IMAGE_ROBOT_2 = 2
IMAGE_ROBOT_3 = 3
IMAGE_ROBOT_4 = 4
IMAGE_ROBOT_5 = 5
IMAGE_HUB_1 = 6
IMAGE_HUB_2 = 7
IMAGE_HUB_3 = 8
IMAGE_HUB_4 = 9
IMAGE_AMUSEMENT_PARK = 10
IMAGE_BEACH = 11
IMAGE_HAUNTED_HOUSE = 12
IMAGE_CARNIVAL = 13
IMAGE_BOOKSHELF = 14
IMAGE_PLAYGROUND = 15
IMAGE_MOON = 16
IMAGE_CAVE = 17
IMAGE_OCEAN = 18
IMAGE_POLAR_BEAR = 19
IMAGE_PARK = 20
IMAGE_RANDOM = 21


def hide() -> None:
    pass


def image(image: int) -> None:
    assert 1 <= image <= 21, "Image must be between 1 and 21"


def show(fullscreen: bool) -> None:
    pass


def text(text: str) -> None:
    pass



"""
The `motor_pair` module is used to run motors in a synchronized fashion. This mode is optimal for creating drivebases where you'd want a pair of motors to start and stop at the same time.

To use the `motor_pair` module simply import the module like so:

>>> import motor_pair

All functions in the module should be called inside the `motor_pair` module as a prefix like so:

>>> motor_pair.pair(motor_pair.PAIR_1, port.A, port.B)
"""

from typing import Awaitable

import motor

PAIR_1 = 0
"""First Motor Pair"""
PAIR_2 = 1
"""Second Motor Pair"""
PAIR_3 = 2
"""Third Motor Pair"""


def move(
    pair: int, steering: int, *, velocity: int = 360, acceleration: int = 1000
) -> None:
    """
    Move a Motor Pair at a constant speed until a new command is given.

    Example
    -------
    ```python
    from hub import port
    import runloop
    import motor_pair

    async def main():
        # Pair motors on port A and B
        motor_pair.pair(motor_pair.PAIR_1, port.A, port.B)

        await runloop.sleep_ms(2000)

        # Move straight at default velocity
        motor_pair.move(motor_pair.PAIR_1, 0)

        await runloop.sleep_ms(2000)

        # Move straight at a specific velocity
        motor_pair.move(motor_pair.PAIR_1, 0, velocity=280)

        await runloop.sleep_ms(2000)

        # Move straight at a specific velocity and acceleration
        motor_pair.move(motor_pair.PAIR_1, 0, velocity=280, acceleration=100)

    runloop.run(main())
    ```

    Parameters
    ----------
    pair : int
        The pair slot of the Motor Pair.
    steering : int
        The steering (-100 to 100) angle to apply (negative values turn left, positive values turn right).
    velocity : int, optional
        The velocity in degrees/sec

        Value ranges depends on motor type.

        - Small motor (essential): -660 to 660
        - Medium motor: -1110 to 1110
        - Large motor: -1050 to 1050
    acceleration : int, optional
        The acceleration (deg/sec²) (1 - 10000)
    """

    assert -100 <= steering <= 100, "Steering must be between -100 and 100"
    assert 1 <= acceleration <= 10_000, (
        "Acceleration must be between 1 and 10000"
    )


async def move_for_degrees(
    pair: int,
    degrees: int,
    steering: int,
    *,
    velocity: int = 360,
    stop: int = motor.BRAKE,
    acceleration: int = 1000,
    deceleration: int = 1000,
) -> Awaitable:
    """
    Move a Motor Pair at a constant speed for a specific number of degrees. When awaited returns a status of the movement that corresponds to one of the following constants from the motor module:

    - `motor.READY`
    - `motor.RUNNING`
    - `motor.STALLED`
    - `motor.CANCELED`
    - `motor.ERROR`
    - `motor.DISCONNECTED`

    Example
    -------
    ```python
    from hub import port
    import runloop
    import motor_pair

    async def main():
        # Pair motors on port A and B
        motor_pair.pair(motor_pair.PAIR_1, port.A, port.B)

        # Move straight at default velocity for 90 degrees
        await motor_pair.move_for_degrees(motor_pair.PAIR_1, 90, 0)

        # Move straight at a specific velocity
        await motor_pair.move_for_degrees(motor_pair.PAIR_1, 360, 0, velocity=280)

        # Move straight at a specific velocity with a slow deceleration
        await motor_pair.move_for_degrees(motor_pair.PAIR_1, 360, 0, velocity=280, deceleration=10)

    runloop.run(main())
    ```

    Parameters
    ---------
    pair : int
        The pair slot of the Motor Pair.
    degrees : int
        The number of degrees
    steering : int
        The steering (-100 to 100)
    velocity : int, optional
        The velocity in degrees/sec

        Value ranges depends on motor type.

        - Small motor (essential): -660 to 660
        - Medium motor: -1110 to 1110
        - Large motor: -1050 to 1050
    stop : int, optional
        The behavior of the Motor after it has stopped. Use the constants in the `motor` module.

        Possible values are
        - `motor.COAST` to make the motor coast until a stop
        - `motor.BRAKE` to brake and continue to brake after stop
        - `motor.HOLD` to tell the motor to hold it's position
        - `motor.CONTINUE` to tell the motor to keep running at whatever velocity it's running at until it gets another command
        - `motor.SMART_COAST` to make the motor brake until stop and then coast and compensate for inaccuracies in the next command
        - `motor.SMART_BRAKE` to make the motor brake and continue to brake after stop and compensate for inaccuracies in the next command
    acceleration : int, optional
        The acceleration (deg/sec²) (1 - 10000)
    deceleration : int, optional
        The deceleration (deg/sec²) (1 - 10000)
    """
    assert -100 <= steering <= 100, "Steering must be between -100 and 100"
    assert 1 <= acceleration <= 10_000, (
        "Acceleration must be between 1 and 10000"
    )
    assert 1 <= deceleration <= 10_000, (
        "Deceleration must be between 1 and 10000"
    )
    assert stop in (
        motor.COAST,
        motor.BRAKE,
        motor.HOLD,
        motor.CONTINUE,
        motor.SMART_COAST,
        motor.SMART_BRAKE,
    ), "Invalid stop value"


async def move_for_time(
    pair: int,
    duration: int,
    steering: int,
    *,
    velocity: int = 360,
    stop: int = motor.BRAKE,
    acceleration: int = 1000,
    deceleration: int = 1000,
) -> Awaitable:
    """
    Move a Motor Pair at a constant speed for a specific duration. When awaited returns a status of the movement that corresponds to one of the following constants from the motor module:

    - `motor.READY`
    - `motor.RUNNING`
    - `motor.STALLED`
    - `motor.CANCELED`
    - `motor.ERROR`
    - `motor.DISCONNECTED`

    Example
    -------
    ```python
    from hub import port
    import runloop
    import motor_pair

    async def main():
        # Pair motors on port A and B
        motor_pair.pair(motor_pair.PAIR_1, port.A, port.B)

        # Move straight at default velocity for 1 second
        await motor_pair.move_for_time(motor_pair.PAIR_1, 1000, 0)

        # Move straight at a specific velocity for 1 second
        await motor_pair.move_for_time(motor_pair.PAIR_1, 1000, 0, velocity=280)

        # Move straight at a specific velocity for 10 seconds with a slow deceleration
        await motor_pair.move_for_time(motor_pair.PAIR_1, 10000, 0, velocity=280, deceleration=10)

    runloop.run(main())
    ```

    Parameters
    ----------
    pair : int
        The pair slot of the Motor Pair.
    duration : int
        The duration in milliseconds.
    steering : int
        The steering (-100 to 100)
    velocity : int
        The velocity in degrees/sec
        Value ranges depends on motor type.

        - Small motor (essential): -660 to 660
        - Medium motor: -1110 to 1110
        - Large motor: -1050 to 1050
    stop : int, optional
        The behavior of the Motor after it has stopped. Use the constants in the `motor` module.

        Possible values are
        - `motor.COAST` to make the motor coast until a stop
        - `motor.BRAKE` to brake and continue to brake after stop
        - `motor.HOLD` to tell the motor to hold it's position
        - `motor.CONTINUE` to tell the motor to keep running at whatever velocity it's running at until it gets another command
        - `motor.SMART_COAST` to make the motor brake until stop and then coast and compensate for inaccuracies in the next command
        - `motor.SMART_BRAKE` to make the motor brake and continue to brake after stop and compensate for inaccuracies in the next command
    acceleration : int, optional
        The acceleration (deg/sec²) (1 - 10000)
    deceleration : int, optional
        The deceleration (deg/sec²) (1 - 10000)
    """
    assert -100 <= steering <= 100, "Steering must be between -100 and 100"
    assert 1 <= acceleration <= 10_000, (
        "Acceleration must be between 1 and 10000"
    )
    assert 1 <= deceleration <= 10_000, (
        "Deceleration must be between 1 and 10000"
    )
    assert stop in (
        motor.COAST,
        motor.BRAKE,
        motor.HOLD,
        motor.CONTINUE,
        motor.SMART_COAST,
        motor.SMART_BRAKE,
    ), "Invalid stop value"


def move_tank(
    pair: int,
    left_velocity: int,
    right_velocity: int,
    *,
    acceleration: int = 1000,
) -> None:
    """
    Perform a tank move on a Motor Pair at a constant speed until a new command is given.

    Example
    -------
    ```python
    from hub import port
    import runloop
    import motor_pair

    async def main():
        # Pair motors on port A and B
        motor_pair.pair(motor_pair.PAIR_1, port.A, port.B)

        # Move straight at default velocity
        motor_pair.move_tank(motor_pair.PAIR_1, 1000, 1000)

        await runloop.sleep_ms(2000)

        # Turn right
        motor_pair.move_tank(motor_pair.PAIR_1, 0, 1000)

        await runloop.sleep_ms(2000)

        # Perform tank turn
        motor_pair.move_tank(motor_pair.PAIR_1, 1000, -1000)

    runloop.run(main())
    ```

    Parameters
    ----------
    pair : int
        The pair slot of the Motor Pair.
    left_velocity : int
        The velocity (deg/sec) of the left motor.
    right_velocity : int
        The velocity (deg/sec) of the right motor.
    acceleration : int, optional
        The acceleration (deg/sec²) (1 - 10000)
    """
    assert 1 <= acceleration <= 10_000, (
        "Acceleration must be between 1 and 10000"
    )


async def move_tank_for_degrees(
    pair: int,
    degrees: int,
    left_velocity: int,
    right_velocity: int,
    *,
    stop: int = motor.BRAKE,
    acceleration: int = 1000,
    deceleration: int = 1000,
) -> Awaitable:
    """
    Perform a tank move on a Motor Pair at a constant speed until a new command is given. When awaited returns a status of the movement that corresponds to one of the following constants from the motor module:

    - `motor.READY`
    - `motor.RUNNING`
    - `motor.STALLED`
    - `motor.CANCELED`
    - `motor.ERROR`
    - `motor.DISCONNECTED`

    Example
    -------
    ```python
    from hub import port
    import runloop
    import motor_pair

    async def main():
        # Pair motors on port A and B
        motor_pair.pair(motor_pair.PAIR_1, port.A, port.B)

        # Move straight at default velocity for 360 degrees
        await motor_pair.move_tank_for_degrees(motor_pair.PAIR_1, 360, 1000, 1000)

        # Turn right for 180 degrees
        await motor_pair.move_tank_for_degrees(motor_pair.PAIR_1, 180, 0, 1000)

        # Perform tank turn for 720 degrees
        await motor_pair.move_tank_for_degrees(motor_pair.PAIR_1, 720, 1000, -1000)

    runloop.run(main())
    ```

    Parameters
    ----------
    pair : int
        The pair slot of the Motor Pair.
    degrees : int
        The number of degrees to rotate the motors.
    left_velocity : int
        The velocity (deg/sec) of the left motor.
    right_velocity : int
        The velocity (deg/sec) of the right motor.
    stop : int, optional
        The behavior of the Motor after it has stopped. Use the constants in the `motor` module.
        Possible values are
        - `motor.COAST` to make the motor coast until a stop
        - `motor.BRAKE` to brake and continue to brake after stop
        - `motor.HOLD` to tell the motor to hold it's position
        - `motor.CONTINUE` to tell the motor to keep running at whatever velocity it's running at until it gets another command
        - `motor.SMART_COAST` to make the motor brake until stop and then coast and compensate for inaccuracies in the next command
        - `motor.SMART_BRAKE` to make the motor brake and continue to brake after stop and compensate for inaccuracies in the next command
    acceleration : int, optional
        The acceleration (deg/sec²) (1 - 10000)
    deceleration : int, optional
        The deceleration (deg/sec²) (1 - 10000)
    """
    assert 1 <= acceleration <= 10_000, (
        "Acceleration must be between 1 and 10000"
    )
    assert 1 <= deceleration <= 10_000, (
        "Deceleration must be between 1 and 10000"
    )
    assert stop in (
        motor.COAST,
        motor.BRAKE,
        motor.HOLD,
        motor.CONTINUE,
        motor.SMART_COAST,
        motor.SMART_BRAKE,
    ), "Invalid stop value"


async def move_tank_for_time(
    pair: int,
    left_velocity: int,
    right_velocity: int,
    duration: int,
    *,
    stop: int = motor.BRAKE,
    acceleration: int = 1000,
    deceleration: int = 1000,
) -> Awaitable:
    """
    Perform a tank move on a Motor Pair at a constant speed for a specific amount of time. When awaited returns a status of the movement that corresponds to one of the following constants from the motor module:

    - `motor.READY`
    - `motor.RUNNING`
    - `motor.STALLED`
    - `motor.CANCELED`
    - `motor.ERROR`
    - `motor.DISCONNECTED`

    Example
    -------
    ```python
    from hub import port
    import runloop
    import motor_pair

    async def main():
        # Pair motors on port A and B
        motor_pair.pair(motor_pair.PAIR_1, port.A, port.B)

        # Move straight at default velocity for 1 second
        await motor_pair.move_tank_for_time(motor_pair.PAIR_1, 1000, 1000, 1000)

        # Turn right for 3 seconds
        await motor_pair.move_tank_for_time(motor_pair.PAIR_1, 0, 1000, 3000)

        # Perform tank turn for 2 seconds
        await motor_pair.move_tank_for_time(motor_pair.PAIR_1, 1000, -1000, 2000)

    runloop.run(main())
    ```

    Parameters
    ----------
    pair : int
        The pair slot of the Motor Pair.
    left_velocity : int
        The velocity (deg/sec) of the left motor.
    right_velocity : int
        The velocity (deg/sec) of the right motor.
    duration : int
        The duration (ms) to run the motors.
    stop : int, optional
        The behavior of the Motor after it has stopped. Use the constants in the `motor` module.
        Possible values are
        - `motor.COAST` to make the motor coast until a stop
        - `motor.BRAKE` to brake and continue to brake after stop
        - `motor.HOLD` to tell the motor to hold it's position
        - `motor.CONTINUE` to tell the motor to keep running at whatever velocity it's running at until it gets another command
        - `motor.SMART_COAST` to make the motor brake until stop and then coast and compensate for inaccuracies in the next command
        - `motor.SMART_BRAKE` to make the motor brake and continue to brake after stop and compensate for inaccuracies in the next command
    acceleration : int, optional
        The acceleration (deg/sec²) (1 - 10000)
    deceleration : int, optional
        The deceleration (deg/sec²) (1 - 10000)
    """
    assert 1 <= acceleration <= 10_000, (
        "Acceleration must be between 1 and 10000"
    )
    assert 1 <= deceleration <= 10_000, (
        "Deceleration must be between 1 and 10000"
    )
    assert stop in (
        motor.COAST,
        motor.BRAKE,
        motor.HOLD,
        motor.CONTINUE,
        motor.SMART_COAST,
        motor.SMART_BRAKE,
    ), "Invalid stop value"


def pair(pair: int, left_motor: int, right_motor: int) -> None:
    """
    Pair two motors (`left_motor` & `right_motor`) and store the paired motors in `pair`. Use `pair` in all subsequent `motor_pair` related function calls.

    Example
    -------
    ```python
    import motor_pair
    from hub import port

    motor_pair.pair(motor_pair.PAIR_1, port.A, port.B)
    ```

    Parameters
    ----------
    pair : int
        The pair slot of the Motor Pair.
    left_motor : int
        The port of the left motor. Use the `port` submodule in the `hub` module.
    right_motor : int
        The port of the right motor. Use the `port` submodule in the `hub` module.
    """
    pass


def stop(pair: int, *, stop: int = motor.BRAKE) -> None:
    """
    Stops a Motor Pair.

    Example
    -------
    ```python
    import motor_pair

    motor_pair.stop(motor_pair.PAIR_1)
    ```

    Parameters
    ---------
    pair : int
        The pair slot of the Motor Pair.
    stop : int
        The behavior of the Motor after it has stopped. Use the constants in the `motor` module.
        Possible values are
        - `motor.COAST` to make the motor coast until a stop
        - `motor.BRAKE` to brake and continue to brake after stop
        - `motor.HOLD` to tell the motor to hold it's position
        - `motor.CONTINUE` to tell the motor to keep running at whatever velocity it's running at until it gets another command
        - `motor.SMART_COAST` to make the motor brake until stop and then coast and compensate for inaccuracies in the next command
        - `motor.SMART_BRAKE` to make the motor brake and continue to brake after stop and compensate for inaccuracies in the next command
    """
    pass


def unpair(pair: int) -> None:
    """
    Unpair a Motor Pair.

    Example
    ------
    ```python
    import motor_pair

    motor_pair.unpair(motor_pair.PAIR_1)
    ```

    Parameters
    ----------
    pair : int
        The pair slot of the Motor Pair.
    """
    pass



"The `color_sensor` module enables you to write code that reacts to specific colors or the intensity of the reflected light."


async def color(port: int) -> int:
    "Returns the color value of the detected color. Use the `color` module to map the color value to a specific color."
    assert port in range(6), "Invalid port"


async def reflection(port: int) -> int:
    "Retrieves the intensity of the reflected light (0-100%)."
    assert port in range(6), "Invalid port"


async def rgbi(port: int) -> tuple[int, int, int, int]:
    "Retrieves the overall color intensity and intensity of red, green and blue."
    assert port in range(6), "Invalid port"



from __future__ import annotations
import asyncio
import logging
from typing import Callable, Awaitable, Optional, Tuple, Type, TypeVar, cast

from bleak import BleakClient, BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from bleak.backends.characteristic import BleakGATTCharacteristic

from ._lib.connection import UUID
from ._lib import cobs
from ._lib.crc import crc
from ._lib.messages import *  # BaseMessage, InfoRequest, InfoResponse, etc.

TM = TypeVar("TM", bound="BaseMessage")

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
    datefmt="%H:%M:%S",
)
logger = logging.getLogger(__name__)


class Spike:
    DEVICE_NOTIFICATION_INTERVAL_MS = 5000

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, timeout: int = 10, slot: int = 0) -> None:
        self.timeout = timeout
        self.slot = slot

        self._device: Optional[BLEDevice] = None
        self._client: Optional[BleakClient] = None
        self._service = None
        self._rx = None
        self._tx = None

        self._stop = asyncio.Event()
        self._info: Optional[InfoResponse] = None
        self._pending: Tuple[int, asyncio.Future] = (-1, asyncio.Future())
        self._notify_cb: Optional[
            Callable[[DeviceNotification], Awaitable[None] | None]
        ] = None

    async def connect(self) -> None:
        self._device = await BleakScanner.find_device_by_filter(
            filterfunc=self._match_service, timeout=self.timeout
        )
        if self._device is None:
            logger.error(
                "No SPIKE hub found. Ensure power, range, and pairing state."
            )
            raise RuntimeError("No SPIKE hub found")

        logger.info("Connecting to SPIKE hub...")
        self._client = BleakClient(
            self._device, disconnected_callback=self._on_disconnect
        )
        await self._client.connect()

        # Some Bleak builds don’t have get_services(); services are available post-connect.
        # If your build does have it, calling it is harmless. Guard with hasattr.
        if hasattr(self._client, "get_services"):
            try:
                await getattr(
                    self._client, "get_services"
                )()  # no-op on some backends
            except TypeError:
                # Older signatures or property-only; ignore
                pass

        svc = self._client.services.get_service(UUID.SERVICE)
        if svc is None:
            raise RuntimeError(
                "SPIKE service not found on the connected device"
            )

        self._service = svc
        self._rx = svc.get_characteristic(UUID.RX)
        self._tx = svc.get_characteristic(UUID.TX)
        if not self._rx or not self._tx:
            raise RuntimeError("SPIKE RX/TX characteristics not found")

        await self._client.start_notify(self._tx, self._on_data)
        logger.info("Connected to SPIKE hub")

    async def disconnect(self) -> None:
        if self._client and self._client.is_connected:
            await self._client.disconnect()
            logger.info("Disconnected from SPIKE hub")

    async def get_info(self) -> InfoResponse:
        logger.info("Requesting hub info...")
        self._info = await self._send_request(InfoRequest(), InfoResponse)
        logger.info(f"Hub info: {self._info}")
        return self._info

    async def enable_notifications(
        self, interval_ms: int = DEVICE_NOTIFICATION_INTERVAL_MS
    ) -> None:
        logger.info(f"Enabling device notifications every {interval_ms} ms")
        resp = await self._send_request(
            DeviceNotificationRequest(interval_ms), DeviceNotificationResponse
        )
        if not resp.success:
            logger.error("Failed to enable notifications")
            raise RuntimeError("Failed to enable notifications")
        logger.info("Device notifications enabled")

    async def clear_slot(self, slot: Optional[int] = None) -> None:
        s = self.slot if slot is None else slot
        logger.info(f"Clearing slot {s}")
        resp = await self._send_request(ClearSlotRequest(s), ClearSlotResponse)
        if not resp.success:
            logger.warning(
                f"Slot {s} not acknowledged as cleared (may already be empty)"
            )

    async def upload_program(
        self,
        program: bytes,
        name: str = "program.py",
        slot: Optional[int] = None,
    ) -> None:
        s = self.slot if slot is None else slot
        info = self._require_info()

        logger.info(f"Starting file upload to slot {s} as '{name}'")
        start = await self._send_request(
            StartFileUploadRequest(name, s, crc(program)),
            StartFileUploadResponse,
        )
        if not start.success:
            logger.error("StartFileUpload not acknowledged")
            raise RuntimeError("StartFileUpload not acknowledged")

        running_crc = 0
        for i in range(0, len(program), info.max_chunk_size):
            chunk = program[i : i + info.max_chunk_size]
            running_crc = crc(chunk, running_crc)
            part = await self._send_request(
                TransferChunkRequest(running_crc, chunk), TransferChunkResponse
            )
            if not part.success:
                logger.error(f"Chunk transfer failed at offset {i}")
                raise RuntimeError(f"Chunk transfer failed at offset {i}")
        logger.info("File upload complete")

    async def start_program(self, slot: Optional[int] = None) -> None:
        s = self.slot if slot is None else slot
        logger.info(f"Starting program in slot {s}")
        resp = await self._send_request(
            ProgramFlowRequest(stop=False, slot=s), ProgramFlowResponse
        )
        if not resp.success:
            logger.error("Failed to start program")
            raise RuntimeError("Failed to start program")
        logger.info("Program started")

    async def run_until_disconnect(self) -> None:
        logger.info("Waiting until disconnect...")
        await self._stop.wait()

    def on_device_notification(
        self, cb: Callable[[DeviceNotification], Awaitable[None] | None]
    ) -> None:
        self._notify_cb = cb

    async def __aenter__(self) -> Spike:
        await self.connect()
        return self

    async def __aexit__(self, _, __, ___) -> None:
        await self.disconnect()

    @staticmethod
    def _match_service(_dev: BLEDevice, adv: AdvertisementData) -> bool:
        uuids = adv.service_uuids or []
        return UUID.SERVICE.lower() in [u.lower() for u in uuids]

    def _on_disconnect(self, _client: BleakClient) -> None:
        logger.warning("SPIKE hub disconnected")
        self._stop.set()

    async def _send(self, msg: BaseMessage) -> None:
        if not self._client or not self._rx:
            raise RuntimeError("Not connected")

        logger.debug(f"Sending: {msg}")
        payload = msg.serialize()
        frame = cobs.pack(payload)
        packet_size = self._info.max_packet_size if self._info else len(frame)

        for i in range(0, len(frame), packet_size):
            await self._client.write_gatt_char(
                self._rx, frame[i : i + packet_size], response=False
            )

    async def _send_request(self, msg: BaseMessage, resp_t: Type[TM]) -> TM:
        loop = asyncio.get_event_loop()
        self._pending = (resp_t.ID, loop.create_future())
        await self._send(msg)
        return cast(TM, await self._pending[1])

    def _on_data(self, _ch: BleakGATTCharacteristic, data: bytearray) -> None:
        if not data or data[-1] != cobs.DELIMITER:
            try:
                un_xor = bytes(b ^ cobs.XOR for b in data)
                logger.warning(f"Incomplete message:\n {un_xor}")
            except Exception:
                logger.warning("Incomplete message (unpack failed)")
            return

        try:
            msg = deserialize(cobs.unpack(data))
            logger.debug(f"Received: {msg}")
        except Exception as e:
            logger.error(f"Decode error: {e}")
            return

        if msg.ID == self._pending[0] and not self._pending[1].done():
            self._pending[1].set_result(msg)

        if isinstance(msg, DeviceNotification):
            if self._notify_cb:
                res = self._notify_cb(msg)
                if asyncio.iscoroutine(res):
                    asyncio.create_task(res)
            else:
                updates = sorted(msg.messages, key=lambda x: x[1])
                for name, value in updates:
                    logger.info(f" - {name:<10}: {value}")

    def _require_info(self) -> InfoResponse:
        if self._info is None:
            raise RuntimeError("Call get_info() first")
        return self._info



"""
To use the Motion Sensor module, add the following import statement to your project:

>>> from hub import motion_sensor

All functions in the module should be called using the `motion_sensor` module as a prefix, for example:

>>> motion_sensor.up_face()
"""

TOP = 0
"""The SPIKE Prime hub face with the Light Matrix."""
FRONT = 1
"""The SPIKE Prime hub face where the speaker is."""
RIGHT = 2
"""The right side of the SPIKE Prime hub when facing the front hub face."""
BOTTOM = 3
"""The side of the SPIKE Prime hub where the battery is."""
BACK = 4
"""The SPIKE Prime hub face with the USB charging port."""
LEFT = 5
"""The left side of the SPIKE Prime hub when facing the front hub face."""


def acceleration(raw_unfiltered: bool) -> tuple[int, int, int]:
    """
    Returns a tuple containing x, y & z acceleration values as integers. The values are mili G, so 1 / 1000 G

    Parameters
    ----------
    raw_unfiltered : bool
        If we want the data back raw and unfiltered
    """
    pass


def angular_velocity(raw_unfiltered: bool) -> tuple[int, int, int]:
    """
    Returns a tuple containing x, y & z angular velocity values as integers. The values are decidegrees per second

    Parameters
    ----------
    raw_unfiltered : bool
        If we want the data back raw and unfiltered
    """
    pass


def gesture() -> int:
    """
    Returns the gesture recognized.

    Possible values are:
    - `motion_sensor.TAPPED`
    - `motion_sensor.DOUBLE_TAPPED`
    - `motion_sensor.SHAKEN`
    - `motion_sensor.FALLING`
    - `motion_sensor.UNKNOWN`
    """
    pass


def get_yaw_face() -> int:
    """
    Retrieve the face of the hub that yaw is relative to.

    If you place the hub on a flat surface with the returned face pointing up, only the yaw value will update as you rotate the hub.

    Returns
    -------
    int
        One of the following constants indicating the hub face:
        - `motion_sensor.TOP`: The SPIKE Prime hub face with the USB charging port.
        - `motion_sensor.FRONT`: The SPIKE Prime hub face with the Light Matrix.
        - `motion_sensor.RIGHT`: The right side of the SPIKE Prime hub when facing the front hub face.
        - `motion_sensor.BOTTOM`: The side of the SPIKE Prime hub where the battery is.
        - `motion_sensor.BACK`: The SPIKE Prime hub face where the speaker is.
        - `motion_sensor.LEFT`: The left side of the SPIKE Prime hub when facing the front hub face.
    """
    pass


def quaternion() -> tuple[float, float, float, float]:
    """Returns the hub orientation quaternion as a tuple[w: float, x: float, y: float, z: float]."""
    pass


def reset_tap_count() -> None:
    """Reset the tap count returned by the `tap_count` function."""
    pass


def reset_yaw(angle: int) -> None:
    """Change the yaw angle offset. The angle set will be the new yaw value."""
    pass


def set_yaw_face(up: int) -> bool:
    """
    Change what hub face is used as the yaw face. If you put the hub on a flat surface with this face pointing up, when you rotate the hub only the yaw will update.

    Parameters
    ----------

    up: int
        The hub face that should be set as the upwards facing hub face.
        Available values are:

        - `motion_sensor.TOP` The SPIKE Prime hub face with the USB charging port.
        - `motion_sensor.FRONT` The SPIKE Prime hub face with the Light Matrix.
        - `motion_sensor.RIGHT` The right side of the SPIKE Prime hub when facing the front hub face.
        - `motion_sensor.BOTTOM` The side of the SPIKE Prime hub where the battery is.
        - `motion_sensor.BACK` The SPIKE Prime hub face where the speaker is.
        - `motion_sensor.LEFT` The left side of the SPIKE Prime hub when facing the front hub face.
    """
    pass


def stable() -> bool:
    """Whether or not the hub is resting flat."""
    pass


def tap_count() -> int:
    """Returns the number of taps recognized since the program started or last time `motion_sensor.reset_tap_count()` was called."""
    pass


def tilt_angles() -> tuple[int, int, int]:
    """Returns a tuple containing yaw pitch and roll values as integers. Values are decidegrees"""
    pass


def up_face() -> int:
    """
    Returns the Hub face that is currently facing up
    - `motion_sensor.TOP` The SPIKE Prime hub face with the USB charging port.
    - `motion_sensor.FRONT` The SPIKE Prime hub face with the Light Matrix.
    - `motion_sensor.RIGHT` The right side of the SPIKE Prime hub when facing the front hub face.
    - `motion_sensor.BOTTOM` The side of the SPIKE Prime hub where the battery is.
    - `motion_sensor.BACK` The SPIKE Prime hub face where the speaker is.
    - `motion_sensor.LEFT` The left side of the SPIKE Prime hub when facing the front hub face."""



"The `device` module enables you to write code to get information about devices plugged into the hub."


async def data(port: int) -> tuple[int]:
    "Retrieve the raw LPF-2 data from a device."
    assert port in range(6), "Invalid port"


async def id(port: int) -> int:
    "Retrieve the device id of a device. Each device has an id based on its type."
    assert port in range(6), "Invalid port"


async def get_duty_cycle(port: int) -> int:
    "Retrieve the duty cycle for a device. Returned values is in range 0 to 10000"
    assert port in range(6), "Invalid port"


async def ready(port: int) -> bool:
    "When a device is attached to the hub it might take a short amount of time before it's ready to accept requests. Use `ready` to test for the readiness of the attached devices."
    assert port in range(6), "Invalid port"


async def set_duty_cycle(port: int, duty_cycle: int) -> None:
    "Set the duty cycle for a device. Accepted values are in range 0 to 10000."
    assert port in range(6), "Invalid port"
    assert duty_cycle in range(10001), "Invalid duty cycle"



"The linegraph module is used make line graphs in the SPIKE App"

from typing import Awaitable


def clear(color: int) -> None:
    assert 0 <= color <= 10, "Color must be between 0 and 10"


def clear_all() -> None:
    pass


async def get_average(color: int) -> Awaitable:
    assert 0 <= color <= 10, "Color must be between 0 and 10"


async def get_last(color: int) -> Awaitable:
    assert 0 <= color <= 10, "Color must be between 0 and 10"


async def get_max(color: int) -> Awaitable:
    assert 0 <= color <= 10, "Color must be between 0 and 10"


async def get_min(color: int) -> Awaitable:
    assert 0 <= color <= 10, "Color must be between 0 and 10"


def hide() -> None:
    pass


def plot(color: int, x: float, y: float) -> None:
    assert 0 <= color <= 10, "Color must be between 0 and 10"


def show(fullscreen: bool) -> None:
    pass



"""
Example implementation of the Consistent Overhead Byte Stuffing (COBS) algorithm
used by the SPIKE™ Prime BLE protocol.

This implementation prioritizes readability and simplicity over performance and
should be used for educational purposes only.
"""

DELIMITER = 0x02
"""Delimiter used to mark end of frame"""

NO_DELIMITER = 0xFF
"""Code word indicating no delimiter in block"""

COBS_CODE_OFFSET = DELIMITER
"""Offset added to code word"""

MAX_BLOCK_SIZE = 84
"""Maximum block size (incl. code word)"""

XOR = 3
"""XOR mask for encoding"""


# Source: https://lego.github.io/spike-prime-docs/encoding.html#cobs-encode
def encode(data: bytes):
    """
    Encode data using COBS algorithm, such that no delimiters are present.
    """
    buffer = bytearray()
    code_index = block = 0

    def begin_block():
        """Append code word to buffer and update code_index and block"""
        nonlocal code_index, block
        code_index = len(buffer)  # index of incomplete code word
        buffer.append(NO_DELIMITER)  # updated later if delimiter is encountered
        block = 1  # no. of bytes in block (incl. code word)

    begin_block()
    for byte in data:
        if byte > DELIMITER:
            # non-delimeter value, write as-is
            buffer.append(byte)
            block += 1

        if byte <= DELIMITER or block > MAX_BLOCK_SIZE:
            # block completed because size limit reached or delimiter found
            if byte <= DELIMITER:
                # reason for block completion is delimiter
                # update code word to reflect block size
                delimiter_base = byte * MAX_BLOCK_SIZE
                block_offset = block + COBS_CODE_OFFSET
                buffer[code_index] = delimiter_base + block_offset
            # begin new block
            begin_block()

    # update final code word
    buffer[code_index] = block + COBS_CODE_OFFSET

    return buffer


# Source: https://lego.github.io/spike-prime-docs/encoding.html#cobs-decode
def decode(data: bytes):
    """
    Decode data using COBS algorithm.
    """
    buffer = bytearray()

    def unescape(code: int):
        """Decode code word, returning value and block size"""
        if code == 0xFF:
            # no delimiter in block
            return None, MAX_BLOCK_SIZE + 1
        value, block = divmod(code - COBS_CODE_OFFSET, MAX_BLOCK_SIZE)
        if block == 0:
            # maximum block size ending with delimiter
            block = MAX_BLOCK_SIZE
            value -= 1
        return value, block

    value, block = unescape(data[0])
    for byte in data[1:]:  # first byte already processed
        block -= 1
        if block > 0:
            buffer.append(byte)
            continue

        # block completed
        if value is not None:
            buffer.append(value)

        value, block = unescape(byte)

    return buffer


# Source: https://lego.github.io/spike-prime-docs/encoding.html#cobs-pack
def pack(data: bytes):
    """
    Encode and frame data for transmission.
    """
    buffer = encode(data)

    # XOR buffer to remove problematic ctrl+C
    for i in range(len(buffer)):
        buffer[i] ^= XOR

    # add delimiter
    buffer.append(DELIMITER)
    return bytes(buffer)


# Source: https://lego.github.io/spike-prime-docs/encoding.html#cobs-unpack
def unpack(frame: bytes):
    """
    Unframe and decode frame.
    """
    start = 0
    if frame[0] == 0x01:  # unused priority byte
        start += 1
    # unframe and XOR
    unframed = bytes(map(lambda x: x ^ XOR, frame[start:-1]))
    return bytes(decode(unframed))



"The color module contains all the color constants to use with the color_matrix, color_sensor and light modules."

BLACK = 0
MAGENTA = 1
PURPLE = 2
BLUE = 3
AZURE = 4
TURQUOISE = 5
GREEN = 6
YELLOW = 7
ORANGE = 8
RED = 9
WHITE = 10
UNKNOWN = -1



# Drum constants
DRUM_SNARE = 1
DRUM_BASS = 2
DRUM_SIDE_STICK = 3
DRUM_CRASH_CYMBAL = 4
DRUM_OPEN_HI_HAT = 5
DRUM_CLOSED_HI_HAT = 6
DRUM_TAMBOURINE = 7
DRUM_HAND_CLAP = 8
DRUM_CLAVES = 9
DRUM_WOOD_BLOCK = 10
DRUM_COWBELL = 11
DRUM_TRIANGLE = 12
DRUM_BONGO = 13
DRUM_CONGA = 14
DRUM_CABASA = 15
DRUM_GUIRO = 16
DRUM_VIBRASLAP = 17
DRUM_CUICA = 18

# Instrument constants
INSTRUMENT_PIANO = 1
INSTRUMENT_ELECTRIC_PIANO = 2
INSTRUMENT_ORGAN = 3
INSTRUMENT_GUITAR = 4
INSTRUMENT_ELECTRIC_GUITAR = 5
INSTRUMENT_BASS = 6
INSTRUMENT_PIZZICATO = 7
INSTRUMENT_CELLO = 8
INSTRUMENT_TROMBONE = 9
INSTRUMENT_CLARINET = 10
INSTRUMENT_SAXOPHONE = 11
INSTRUMENT_FLUTE = 12
INSTRUMENT_WOODEN_FLUTE = 13
INSTRUMENT_BASSOON = 14
INSTRUMENT_CHOIR = 15
INSTRUMENT_VIBRAPHONE = 16
INSTRUMENT_MUSIC_BOX = 17
INSTRUMENT_STEEL_DRUM = 18
INSTRUMENT_MARIMBA = 19
INSTRUMENT_SYNTH_LEAD = 20
INSTRUMENT_SYNTH_PAD = 21


def play_drum(drum: int) -> None:
    assert 1 <= drum <= 18, "Drum must be between 1 and 18"


def play_instrument(instrument: int, note: int, duration: int) -> None:
    assert 1 <= instrument <= 21, "Instrument must be between 1 and 21"
    assert 1 <= note <= 130, "Note must be between 1 and 130"



# Source: https://github.com/LEGO/spike-prime-docs/blob/main/examples/python/messages.py

from __future__ import annotations
from abc import ABC
import struct


class BaseMessage(ABC):
    @property
    def ID(cls) -> int:
        raise NotImplementedError

    def serialize(self) -> bytes:
        raise NotImplementedError

    @staticmethod
    def deserialize(data: bytes) -> BaseMessage:
        raise NotImplementedError

    def __str__(self) -> str:
        props = vars(self)
        plist = ", ".join(f"{k}={v}" for k, v in props.items())
        return f"{self.__class__.__name__}({plist})"


def StatusResponse(name: str, id: int):
    class _StatusResponse(BaseMessage):
        ID = id

        def __init__(self, success: bool):
            self.success = success

        @staticmethod
        def deserialize(data: bytes):
            _, status = struct.unpack("<BB", data)
            return _StatusResponse(status == 0x00)

    _StatusResponse.__name__ = name
    return _StatusResponse


class InfoRequest(BaseMessage):
    ID = 0x00

    def serialize(self):
        return b"\0"


class InfoResponse(BaseMessage):
    ID = 0x01

    def __init__(
        self,
        rpc_major,
        rpc_minor,
        rpc_build,
        firmware_major,
        firmware_minor,
        firmware_build,
        max_packet_size,
        max_message_size,
        max_chunk_size,
        product_group_device,
    ):
        self.rpc_major = rpc_major
        self.rpc_minor = rpc_minor
        self.rpc_build = rpc_build
        self.firmware_major = firmware_major
        self.firmware_minor = firmware_minor
        self.firmware_build = firmware_build
        self.max_packet_size = max_packet_size
        self.max_message_size = max_message_size
        self.max_chunk_size = max_chunk_size
        self.product_group_device = product_group_device

    @staticmethod
    def deserialize(data: bytes) -> InfoResponse:
        (
            _,
            rpc_major,
            rpc_minor,
            rpc_build,
            firmware_major,
            firmware_minor,
            firmware_build,
            max_packet_size,
            max_message_size,
            max_chunk_size,
            product_group_device,
        ) = struct.unpack("<BBBHBBHHHHH", data)
        return InfoResponse(
            rpc_major,
            rpc_minor,
            rpc_build,
            firmware_major,
            firmware_minor,
            firmware_build,
            max_packet_size,
            max_message_size,
            max_chunk_size,
            product_group_device,
        )


class ClearSlotRequest(BaseMessage):
    ID = 0x46

    def __init__(self, slot: int):
        self.slot = slot

    def serialize(self):
        return struct.pack("<BB", self.ID, self.slot)


ClearSlotResponse = StatusResponse("ClearSlotResponse", 0x47)


class StartFileUploadRequest(BaseMessage):
    ID = 0x0C

    def __init__(self, file_name: str, slot: int, crc: int):
        self.file_name = file_name
        self.slot = slot
        self.crc = crc

    def serialize(self):
        encoded_name = self.file_name.encode("utf8")
        if len(encoded_name) > 31:
            raise ValueError(
                f"UTF-8 encoded file name too long: {len(encoded_name)} +1 >= 32"
            )
        fmt = f"<B{len(encoded_name) + 1}sBI"
        return struct.pack(fmt, self.ID, encoded_name, self.slot, self.crc)


StartFileUploadResponse = StatusResponse("StartFileUploadResponse", 0x0D)


class TransferChunkRequest(BaseMessage):
    ID = 0x10

    def __init__(self, running_crc: int, chunk: bytes):
        self.running_crc = running_crc
        self.size = len(chunk)
        self.payload = chunk

    def serialize(self):
        fmt = f"<BIH{self.size}s"
        return struct.pack(
            fmt, self.ID, self.running_crc, self.size, self.payload
        )


TransferChunkResponse = StatusResponse("TransferChunkResponse", 0x11)


class ProgramFlowRequest(BaseMessage):
    ID = 0x1E

    def __init__(self, stop: bool, slot: int):
        self.stop = stop
        self.slot = slot

    def serialize(self):
        return struct.pack("<BBB", self.ID, self.stop, self.slot)


ProgramFlowResponse = StatusResponse("ProgramFlowResponse", 0x1F)


class ProgramFlowNotification(BaseMessage):
    ID = 0x20

    def __init__(self, stop: bool):
        self.stop = stop

    @staticmethod
    def deserialize(data: bytes) -> ProgramFlowNotification:
        _, stop = struct.unpack("<BB", data)
        return ProgramFlowNotification(bool(stop))


class ConsoleNotification(BaseMessage):
    ID = 0x21

    def __init__(self, text: str):
        self.text = text

    @staticmethod
    def deserialize(data: bytes) -> ConsoleNotification:
        text_bytes = data[1:].rstrip(b"\0")
        return ConsoleNotification(text_bytes.decode("utf8"))

    def __str__(self):
        return f"{self.__class__.__name__}({self.text!r})"


class DeviceNotificationRequest(BaseMessage):
    ID = 0x28

    def __init__(self, interval_ms: int):
        self.interval_ms = interval_ms

    def serialize(self):
        return struct.pack("<BH", self.ID, self.interval_ms)


DeviceNotificationResponse = StatusResponse("DeviceNotificationResponse", 0x29)

DEVICE_MESSAGE_MAP = {
    0x00: ("Battery", "<BB"),
    0x01: ("IMU", "<BBBhhhhhhhhh"),
    0x02: ("5x5", "<B25B"),
    0x0A: ("Motor", "<BBBhhbi"),
    0x0B: ("Force", "<BBBB"),
    0x0C: ("Color", "<BBbHHH"),
    0x0D: ("Distance", "<BBh"),
    0x0E: ("3x3", "<BB9B"),
}


class DeviceNotification(BaseMessage):
    ID = 0x3C

    def __init__(self, size: int, payload: bytes):
        self.size = size
        self._payload = payload
        self.messages = []
        self.unknown_id: int | None = None
        self.raw_tail: bytes | None = None

        data = payload[:]
        while data:
            msg_id = data[0]
            spec = DEVICE_MESSAGE_MAP.get(msg_id)
            if not spec:
                self.unknown_id = msg_id
                self.raw_tail = data  # preserve for reverse-engineering
                # print(f"Unknown message: 0x{msg_id:02X}, tail={data[:16].hex(' ')}")
                break

            name, fmt = spec
            sz = struct.calcsize(fmt)
            if len(data) < sz:
                # truncated; stop cleanly
                self.raw_tail = data
                break

            values = struct.unpack(fmt, data[:sz])
            self.messages.append((name, values))
            data = data[sz:]

    @staticmethod
    def deserialize(data: bytes) -> DeviceNotification:
        _, size = struct.unpack("<BH", data[:3])
        if len(data) != size + 3:
            print(f"Unexpected size: {len(data)} != {size} + 3")
        return DeviceNotification(size, data[3:])

    def __str__(self):
        return f"{self.__class__.__name__}({[x[0] for x in self.messages]})"


KNOWN_MESSAGES = {
    M.ID: M
    for M in (
        InfoRequest,
        InfoResponse,
        ClearSlotRequest,
        ClearSlotResponse,
        StartFileUploadRequest,
        StartFileUploadResponse,
        TransferChunkRequest,
        TransferChunkResponse,
        ProgramFlowRequest,
        ProgramFlowResponse,
        ProgramFlowNotification,
        ConsoleNotification,
        DeviceNotificationRequest,
        DeviceNotificationResponse,
        DeviceNotification,
    )
}


def deserialize(data: bytes):
    message_type = data[0]
    if message_type in KNOWN_MESSAGES:
        return KNOWN_MESSAGES[message_type].deserialize(data)
    raise ValueError(f"Unknown message: {data.hex(' ')}")



def clear(port: int) -> None:
    "Turn off all pixels on a Color Matrix"
    assert port in range(6), "Invalid port"


def get_pixel(port: int, x: int, y: int) -> tuple[int, int]:
    "Retrieve a specific pixel represented as a tuple containing the color and intensity"
    assert port in range(6), "Invalid port"


def set_pixel(port: int, x: int, y: int, pixel: tuple[int, int]) -> None:
    "Change a single pixel on a Color Matrix"
    assert port in range(6), "Invalid port"


def show(port: int, pixels: list[tuple[int, int]]) -> None:
    "Change all pixels at once on a Color Matrix"
    assert port in range(6), "Invalid port"



"""
To use the Sound module add the following import statement to your project:

>>> from hub import sound

All functions in the module should be called inside the `sound` module as a prefix like so:

>>> sound.stop()

"""

from typing import Awaitable

ANY = -2
DEFAULT = -1
WAVEFORM_SINE = 1
WAVEFORM_SAWTOOTH = 3
WAVEFORM_SQUARE = 2
WAVEFORM_TRIANGLE = 1


async def beep(
    freq: int = 440,
    duration: int = 500,
    volume: int = 100,
    *,
    attack: int = 0,
    decay: int = 0,
    sustain: int = 100,
    release: int = 0,
    transition: int = 10,
    waveform: int = WAVEFORM_SINE,
    channel: int = DEFAULT,
) -> Awaitable:
    """
    Plays a beep sound from the hub

    Parameters:
        freq: int
            The frequency to play
        duration: int
            The duration in milliseconds
        volume: int
            The volume (0-100)
        attack: int
            The time taken for initial run-up of level from nil to peak, beginning when the key is pressed.
        decay: int
            The time taken for the subsequent run down from the attack level to the designated sustain level.
        sustain: int
            The level during the main sequence of the sound's duration, until the key is released.
        release: int
            The time taken for the level to decay from the sustain level to zero after the key is released
        transition: int
            The time in milliseconds to transition into the sound if something is already playing in the channel
        waveform: int
            The synthesized waveform. Use one of the constants in the `hub.sound` module.
        channel: int
            The desired channel to play on, options are `sound.DEFAULT` and `sound.ANY`
    """
    pass


def stop() -> None:
    """Stops all noise from the hub"""
    pass


def volume(volume: int) -> None:
    """
    Parameters
    ----------
        volume: int
            The volume (0-100)
    """



from typing import Callable

async def run_fn(
    program: Callable[[], None],
    *,
    slot: int = 0,
    stay_connected: bool = False,
):
    """Run a Python function on the SPIKE Prime hub."""
    from .spike import logger
    from ._utils import fn_to_string
    try:
        # call function and see if any front-side errors trigger
        program()
    except Exception as e:
        logger.error(f"Error occurred while running program: {e}")
    await run_str(
        fn_to_string(program),
        slot=slot,
        name=program.__name__ + ".py",
        stay_connected=stay_connected
    )

run = run_fn # shorthand alias default

async def run_file(
    program_path: str,
    *,
    slot: int = 0,
    stay_connected: bool = False,
):
    """Run a Python file on the SPIKE Prime hub."""
    from .spike import logger
    from pathlib import Path
    if not Path(program_path).exists():
        logger.error(f"File not found: {program_path}")
        return

    program_str = Path(program_path).read_text()
    try:
        compile(program_str, program_path, "exec")
    except Exception as e:
        logger.error(f"Error compiling {program_path}: {e}")
        return
    await run_str(
        program_str,
        slot=slot,
        name=Path(program_path).name,
        stay_connected=stay_connected
    )

async def run_str(
    program_str: str,
    *,
    slot: int = 0,
    name: str = "program.py",
    stay_connected: bool = False,
):
    """Run a Python string as code on the SPIKE Prime hub."""
    from .spike import Spike
    async with Spike(timeout=10, slot=slot) as hub:
        await hub.get_info()
        await hub.enable_notifications()
        await hub.clear_slot()
        await hub.upload_program(program_str.encode("utf-8"), name=name)
        await hub.start_program()
        if stay_connected:
            try:
                await hub.run_until_disconnect()
            except KeyboardInterrupt:
                await hub.disconnect()


"""
This module contains constants that enables easy access to the ports on the SPIKE Prime hub. Use the constants in all functions that takes a `port` parameter.

To use the Port module add the following import statement to your project:

>>> from hub import port

All functions in the module should be called inside the port module as a prefix like so:

>>> port.A

"""

A = 0
"""The Port that is labelled 'A' on the Hub."""
B = 1
"""The Port that is labelled 'B' on the Hub."""
C = 2
"""The Port that is labelled 'C' on the Hub."""
D = 3
"""The Port that is labelled 'D' on the Hub."""
E = 4
"""The Port that is labelled 'E' on the Hub."""
F = 5
"""The Port that is labelled 'F' on the Hub."""



"""
To use the Light Matrix module add the following import statement to your project:

>>> from hub import light_matrix

All functions in the module should be called inside the light_matrix module as a prefix like so:

>>> light_matrix.write("Hello World")
"""

from typing import Awaitable

IMAGE_HEART = 1
IMAGE_HEART_SMALL = 2
IMAGE_HAPPY = 3
IMAGE_SMILE = 4
IMAGE_SAD = 5
IMAGE_CONFUSED = 6
IMAGE_ANGRY = 7
IMAGE_ASLEEP = 8
IMAGE_SURPRISED = 9
IMAGE_SILLY = 10
IMAGE_FABULOUS = 11
IMAGE_MEH = 12
IMAGE_YES = 13
IMAGE_NO = 14
IMAGE_CLOCK12 = 15
IMAGE_CLOCK1 = 16
IMAGE_CLOCK2 = 17
IMAGE_CLOCK3 = 18
IMAGE_CLOCK4 = 19
IMAGE_CLOCK5 = 20
IMAGE_CLOCK6 = 21
IMAGE_CLOCK7 = 22
IMAGE_CLOCK8 = 23
IMAGE_CLOCK9 = 24
IMAGE_CLOCK10 = 25
IMAGE_CLOCK11 = 26
IMAGE_ARROW_N = 27
IMAGE_ARROW_NE = 28
IMAGE_ARROW_E = 29
IMAGE_ARROW_SE = 30
IMAGE_ARROW_S = 31
IMAGE_ARROW_SW = 32
IMAGE_ARROW_W = 33
IMAGE_ARROW_NW = 34
IMAGE_GO_RIGHT = 35
IMAGE_GO_LEFT = 36
IMAGE_GO_UP = 37
IMAGE_GO_DOWN = 38
IMAGE_TRIANGLE = 39
IMAGE_TRIANGLE_LEFT = 40
IMAGE_CHESSBOARD = 41
IMAGE_DIAMOND = 42
IMAGE_DIAMOND_SMALL = 43
IMAGE_SQUARE = 44
IMAGE_SQUARE_SMALL = 45
IMAGE_RABBIT = 46
IMAGE_COW = 47
IMAGE_MUSIC_CROTCHET = 48
IMAGE_MUSIC_QUAVER = 49
IMAGE_MUSIC_QUAVERS = 50
IMAGE_PITCHFORK = 51
IMAGE_XMAS = 52
IMAGE_PACMAN = 53
IMAGE_TARGET = 54
IMAGE_TSHIRT = 55
IMAGE_ROLLERSKATE = 56
IMAGE_DUCK = 57
IMAGE_HOUSE = 58
IMAGE_TORTOISE = 59
IMAGE_BUTTERFLY = 60
IMAGE_STICKFIGURE = 61
IMAGE_GHOST = 62
IMAGE_SWORD = 63
IMAGE_GIRAFFE = 64
IMAGE_SKULL = 65
IMAGE_UMBRELLA = 66
IMAGE_SNAKE = 67


def clear() -> None:
    """
    Switches off all of the pixels on the Light Matrix.

    Example
    -------
    ```python

    from hub import light_matrix
    import time
    # Update pixels to show an image on Light Matrix, and then turn them off using the clear function

    # Show a small heart
    light_matrix.show_image(2)

    # Wait for two seconds
    time.sleep_ms(2000)

    # Switch off the heart
    light_matrix.clear()
    ```
    """
    pass


def get_orientation() -> int:
    """
    Retrieve the current orientation of the Light Matrix.
    Can be used with the following constants: `orientation.UP`, `orientation.LEFT`, `orientation.RIGHT`, `orientation.DOWN`
    """
    pass


def get_pixel(x: int, y: int) -> int:
    """
    Retrieve the intensity of a specific pixel on the Light Matrix.

    Example
    -------
    ```python

    from hub import light_matrix

    # Show a heart
    light_matrix.show_image(1)

    # Print the value of the center pixel's intensity
    print(light_matrix.get_pixel(2, 2))
    ```

    Parameters
    ----------

    x : int
        The X value, range (0 - 4)
    y : int
        The Y value, range (0 - 4)
    """
    pass


def set_orientation(top: int) -> int:
    """
        Change the orientation of the Light Matrix. All subsequent calls will use the new orientation.
    Can be used with the following constants: `orientation.UP`, `orientation.LEFT`, `orientation.RIGHT`, `orientation.DOWN`

        Parameters
        ----------
        top : int
            The side of the hub to be the top
    """
    pass


def set_pixel(x: int, y: int, intensity: int) -> None:
    """
    Sets the brightness of one pixel (one of the 25 LEDs) on the Light Matrix.

    Example
    -------
    ```python
    from hub import light_matrix
    # Turn on the pixel in the center of the hub
    light_matrix.set_pixel(2, 2, 100)
    ```

    Parameters
    ----------
    x : int
        The X value, range (0 - 4)
    y : int
        The Y value, range (0 - 4)
    intensity : int
        How bright to light up the pixel
    """
    pass


def show(pixels: list[int]) -> None:
    """
    Change all the lights at the same time.

    Example
    -------
    ```python
    from hub import light_matrix
    # Update all pixels on Light Matrix using the show function

    # Create a list with 25 identical intensity values
    pixels = [100] * 25

    # Update all pixels to show same intensity
    light_matrix.show(pixels)
    ```

    Parameters
    ----------
    pixels : list[int]
        A list containing light intensity values for all 25 pixels.
    """
    pass


def show_image(image: int) -> None:
    """
    Display one of the built in images on the display.

    Example
    -------
    ```python
    from hub import light_matrix
    # Update pixels to show an image on Light Matrix using the show_image function

    # Show a smiling face
    light_matrix.show_image(light_matrix.IMAGE_HAPPY)
    ```

    Parameters
    ----------
    image : int
        The id of the image to show. The range of available images is 1 to 67. There are consts on the `light_matrix` module for these.
    """
    pass


async def write(
    text: str, intensity: int = 100, time_per_character: int = 500
) -> Awaitable:
    """
    Displays text on the Light Matrix, one letter at a time, scrolling from right to left except if there is a single character to show which will not scroll

    Example
    -------
    ```python
    from hub import light_matrix
    # Write a message to the hub
    light_matrix.write("Hello, world!")
    ```

    Parameters
    ----------
    text : str
        The text to display
    intensity : int
        How bright to light up the pixel
    time_per_character : int
        How long to show each character on the display
    """



class UUID:
    """
    The LEGO SPIKE Prime Hub exposes a BLE GATT service with two characteristics:
    - RX: for receiving data (from the hub's perspective)
    - TX: for transmitting data (from the hub's perspective)
    """

    _BASE = "0000FD02"
    _TAIL = "1000-8000-00805F9B34FB"
    SERVICE = f"{_BASE}-0000-{_TAIL}"
    RX = f"{_BASE}-0001-{_TAIL}"
    TX = f"{_BASE}-0002-{_TAIL}"


class Hardware:
    MAC_ADDR = "3C:E4:B0:AB:D3:3A"


class Name:
    HINTS = ["SPIKE", "Spike", "Prime", "Hub", "Lego"] + [
        str(i) for i in range(1, 100)
    ]



"""The distance_sensor module enables you to write code that reacts to specific distances or light up the Distance Sensor in different ways."""


async def clear(port: int) -> None:
    "Turns off all the lights in the Distance Sensor connected to `port`."
    assert port in range(6), "Invalid port"


async def distance(port: int) -> int:
    "Retrieve the distance in millimeters captured by the Distance Sensor connected to `port`. If the Distance Sensor cannot read a valid distance it will return -1."
    assert port in range(6), "Invalid port"


async def get_pixel(port: int, x: int, y: int) -> int:
    "Retrieve the intensity of a specific light on the Distance Sensor connected to `port`."
    assert port in range(6), "Invalid port"
    assert x in range(4), "Invalid x coordinate"
    assert y in range(4), "Invalid y coordinate"


async def set_pixel(port: int, x: int, y: int, intensity: int) -> None:
    "Changes the intensity of a specific light on the Distance Sensor connected to `port`."
    assert port in range(6), "Invalid port"
    assert x in range(4), "Invalid x coordinate"
    assert y in range(4), "Invalid y coordinate"


async def show(port: int, pixels: list[int]) -> None:
    "Change all the lights at the same time."
    assert port in range(6), "Invalid port"



"""
The light module includes functions to change the color of the light on the SPIKE Prime hub.

To use the Light module add the following import statement to your project:

>>> from hub import light

All functions in the module should be called inside the light module as a prefix like so:

>>> light.color(color.RED)
"""

POWER = 0
"""The power button. On SPIKE Prime it's the button between the left and right buttons."""
CONNECT = 1
"""The light around the Bluetooth connect button on SPIKE Prime."""


def color(light: int, color: int) -> None:
    """
    Change the color of a light on the hub.
    ```python
    from hub import light
    import color

    # Change the light to red
    light.color(light.POWER, color.RED)
    ```

    Parameters
    ----------

    light : int
        The light on the hub
    color : int
        A color from the `color` module
    """
    assert 0 <= color <= 10, "Color must be between 0 and 10"



"""
The `orientation` module contains all the orientation constants to use with the `light_matrix` module.

To use the `orientation` module add the following import statement to your project:

>>> import orientation
"""

UP = 0
RIGHT = 1
DOWN = 2
LEFT = 3



"""
This module allows you to react to buttons being pressed on the hub. You must first import the `button` module to use the buttons.

To use the Button module add the following import statement to your project:


>>> from hub import button

All functions in the module should be called inside the `button` module as a prefix like so:

>>> button.pressed(button.LEFT)

"""

LEFT = 1
"""Left button next to the power button on the SPIKE Prime hub"""
RIGHT = 2
"""Right button next to the power button on the SPIKE Prime hub"""


def pressed(button: int) -> int:
    """
    ```python
    from hub import button
    left_button_press_duration = 0

    # Wait for the left button to be pressed
    while not button.pressed(button.LEFT):
        pass

    # As long as the left button is being pressed, update the `left_button_press_duration` variable
    while button.pressed(button.LEFT):
        left_button_press_duration = button.pressed(button.LEFT)

    print("Left button was pressed for " + str(left_button_press_duration) + " milliseconds")
    ```
    """
